"""Icinga module tests."""
import logging
import re
import shlex
from datetime import timedelta
from typing import Sequence, Tuple
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet
from cumin.transports import Command

from spicerack import icinga
from spicerack.administrative import Reason
from spicerack.remote import RemoteHosts
from spicerack.tests import get_fixture_path


def set_mocked_icinga_host_output(mocked_icinga_host, output, n_hosts=1):
    """Setup the mocked icinga_host return value with the given output."""
    out = MsgTreeElem(output.encode(), parent=MsgTreeElem())
    mocked_icinga_host.run_sync.return_value = iter(n_hosts * [(NodeSet("icinga-host"), out)])


def set_mocked_icinga_host_outputs(mocked_icinga_host, outputs):
    """Setup the mocked icinga_host side effect to return the given outputs, one after another."""
    outs = [MsgTreeElem(output.encode(), parent=MsgTreeElem()) for output in outputs]
    mocked_icinga_host.run_sync.side_effect = [iter([(NodeSet("icinga-host"), out)]) for out in outs]


def assert_has_downtime_calls(
    mocked_icinga_host: mock.MagicMock,
    hosts: Sequence[str],
    reason: Reason,
    start: int = 1514764800,
    duration: int = 14400,
):
    """Assert that the mocked icinga_host was called correctly to downtime the given hosts."""
    end = start + duration
    args = f"{start};{end};1;0;{duration};{reason.owner};{reason.reason}"
    downtime_calls = [
        [
            "bash -c "
            + shlex.quote(f'echo -n "[{start}] SCHEDULE_HOST_DOWNTIME;{host};{args}" > /var/lib/icinga/rw/icinga.cmd ')
            for host in hosts
        ],
        [
            "bash -c "
            + shlex.quote(
                f'echo -n "[{start}] SCHEDULE_HOST_SVC_DOWNTIME;{host};{args}" > /var/lib/icinga/rw/icinga.cmd '
            )
            for host in hosts
        ],
    ]
    mocked_icinga_host.run_sync.assert_has_calls(
        [mock.call(*call, print_output=False, print_progress_bars=False) for call in downtime_calls]
    )


def assert_has_service_downtime_calls(
    mocked_icinga_host: mock.MagicMock,
    host_services: Sequence[Tuple[str, str]],
    reason: Reason,
    start: int = 1514764800,
    duration: int = 14400,
):
    """Assert that the mocked icinga_host was called correctly to downtime the given services."""
    end = start + duration
    args = f"{start};{end};1;0;{duration};{reason.owner};{reason.reason}"
    downtime_calls = [
        [
            f'bash -c \'echo -n "[{start}] SCHEDULE_SVC_DOWNTIME;{host};{service};{args}" '
            f"> /var/lib/icinga/rw/icinga.cmd '"
            for host, service in host_services
        ],
    ]
    mocked_icinga_host.run_sync.assert_has_calls(
        [mock.call(*call, print_output=False, print_progress_bars=False) for call in downtime_calls]
    )


class TestCommandFile:
    """Test class for the CommandFile class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_icinga_host.__len__.return_value = 1
        set_mocked_icinga_host_output(self.mocked_icinga_host, "/var/lib/icinga/rw/icinga.cmd")

    @pytest.mark.parametrize("num_hosts", (0, 2))
    def test_wrong_hosts(self, num_hosts):
        """It should raise IcingaError if no or too many hosts match the RemoteHosts instance."""
        self.mocked_icinga_host.__len__.return_value = num_hosts
        with pytest.raises(icinga.IcingaError, match="Icinga host must match a single host"):
            icinga.CommandFile(self.mocked_icinga_host)

        self.mocked_icinga_host.run_sync.assert_not_called()

    def test_uncached(self):
        """It should return the command_file setting from the Icinga configuration on the Icinga host."""
        assert icinga.CommandFile(self.mocked_icinga_host) == "/var/lib/icinga/rw/icinga.cmd"
        self.mocked_icinga_host.run_sync.assert_called_once()

    @pytest.mark.parametrize(
        "output",
        (
            "",
            " ",
        ),
    )
    def test_no_config(self, output):
        """It should raise IcingaError if failing to get the configuration value."""
        set_mocked_icinga_host_output(self.mocked_icinga_host, output)
        with pytest.raises(icinga.IcingaError, match="Unable to read command_file configuration"):
            icinga.CommandFile(self.mocked_icinga_host)

    def test_cached(self):
        """It should return the already cached value of the command_file if accessed again."""
        command_file = icinga.CommandFile(self.mocked_icinga_host)
        self.mocked_icinga_host.reset_mock()
        assert icinga.CommandFile(self.mocked_icinga_host) == command_file
        self.mocked_icinga_host.assert_not_called()

    def test_uncached_with_other_cached(self):
        """It should get the new command file if the host changes even if has another value already cached."""
        mocked_other_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        mocked_other_icinga_host.__len__.return_value = 1
        set_mocked_icinga_host_output(mocked_other_icinga_host, "/var/lib/icinga/rw/icinga_other.cmd")

        icinga.CommandFile(self.mocked_icinga_host)
        self.mocked_icinga_host.reset_mock()

        assert icinga.CommandFile(mocked_other_icinga_host) == "/var/lib/icinga/rw/icinga_other.cmd"
        self.mocked_icinga_host.run_sync.assert_not_called()
        mocked_other_icinga_host.run_sync.assert_called_once()


def test_status_not_found_error_one_host():
    """Test initializing IcingaStatusNotFoundError with a single host."""
    e = icinga.IcingaStatusNotFoundError(["host1"])
    assert str(e) == "Host host1 was not found in Icinga status"


def test_status_not_found_error_multiple_hosts():
    """Test initializing IcingaStatusNotFoundError with multiple hosts."""
    e = icinga.IcingaStatusNotFoundError(["host1", "host2", "host3"])
    assert str(e) == "Hosts host1, host2, host3 were not found in Icinga status"


class TestIcingaHosts:
    """Test class for the IcingaHosts class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.reason = Reason("Downtime reason", "user1", "orchestration-host", task_id="T12345")
        self.mocked_icinga_host = mock.MagicMock(spec_set=RemoteHosts)
        self.mocked_icinga_host.__len__.return_value = 1
        self.hosts = ["host1.example.com"]
        set_mocked_icinga_host_output(self.mocked_icinga_host, "/var/lib/icinga/rw/icinga.cmd")
        self.icinga_hosts = icinga.IcingaHosts(self.mocked_icinga_host, self.hosts)
        self.mocked_icinga_host.reset_mock()

    def test_init_no_hosts(self):
        """It should raise IcingaError if there are no target hosts."""
        with pytest.raises(icinga.IcingaError, match="Got empty target hosts list"):
            icinga.IcingaHosts(self.mocked_icinga_host, [])

    @pytest.mark.parametrize(
        "target_hosts, verbatim_hosts, effective_hosts",
        (
            (["host1.example.com", "host2.example.com"], False, NodeSet("host[1-2]")),
            (["host1.example.com", "host2.example.com"], True, NodeSet("host[1-2].example.com")),
            (NodeSet("host1.example.com"), False, NodeSet("host1")),
            (NodeSet("host1.example.com"), True, NodeSet("host1.example.com")),
        ),
    )
    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_init_hosts(self, mocked_time, target_hosts, verbatim_hosts, effective_hosts):
        """It should init the target hosts according to the input and verbatim option."""
        instance = icinga.IcingaHosts(self.mocked_icinga_host, target_hosts, verbatim_hosts=verbatim_hosts)
        instance.run_icinga_command("TEST_COMMAND", "arg1", "arg2")
        calls = [
            f"bash -c 'echo -n \"[1514764800] TEST_COMMAND;{host};arg1;arg2\" > /var/lib/icinga/rw/icinga.cmd '"
            for host in effective_hosts
        ]

        self.mocked_icinga_host.run_sync.assert_called_once_with(*calls, print_output=False, print_progress_bars=False)
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtimed(self, mocked_time):
        """It should downtime the hosts on Icinga, yield and delete the downtime once done."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with self.icinga_hosts.downtimed(self.reason):
            assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], self.reason)
            self.mocked_icinga_host.run_sync.reset_mock()

        self.mocked_icinga_host.run_sync.assert_has_calls(
            [
                mock.call(
                    'bash -c \'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1" > '
                    "/var/lib/icinga/rw/icinga.cmd '",
                    print_output=False,
                    print_progress_bars=False,
                )
            ]
        )
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_is_kept_when_exception_is_raised(self, _mocked_time):
        """Downtime should not be removed if an exception is raised."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with pytest.raises(ValueError):
            with self.icinga_hosts.downtimed(self.reason):
                assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], self.reason)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert not self.mocked_icinga_host.run_sync.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_is_removed_when_exception_is_raised(self, _mocked_time):
        """Downtime should not be removed if an exception is raised."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with pytest.raises(ValueError):
            with self.icinga_hosts.downtimed(self.reason, remove_on_error=True):
                assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], self.reason)
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert self.mocked_icinga_host.run_sync.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_default_params(self, _mocked_time):
        """It should downtime the hosts on the Icinga server with the default params."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.downtime(self.reason)
        assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], self.reason)

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_with_apostrophe_in_reason(self, _mocked_time):
        """It should correctly quote the apostrophe in the reason string."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        reason = Reason("An apostrophe's here", "user1", "orchestration-host", task_id="T12345")
        self.icinga_hosts.downtime(reason)
        assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], reason)

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_custom_duration(self, _mocked_time):
        """It should downtime the hosts for the given duration on the Icinga server."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.downtime(self.reason, duration=timedelta(minutes=30))
        assert_has_downtime_calls(self.mocked_icinga_host, ["host1"], self.reason, duration=1800)

    def test_downtime_invalid_duration(self):
        """It should raise IcingaError if the duration is too short."""
        with pytest.raises(icinga.IcingaError, match="Downtime duration must be at least 1 minute"):
            self.icinga_hosts.downtime(self.reason, duration=timedelta(seconds=59))

    def test_downtime_unknown_host(self):
        """It should raise IcingaError if any host is not known to Icinga."""
        set_mocked_icinga_host_output(self.mocked_icinga_host, '{"host1": null}')
        with pytest.raises(
            icinga.IcingaError, match="Host host1 was not found in Icinga status - no hosts have been downtimed"
        ):
            self.icinga_hosts.downtime(self.reason)

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_services_downtimed(self, mocked_time):
        """It should downtime the hosts on Icinga, yield and delete the downtime once done."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as before:
            with open(get_fixture_path("icinga", "status_with_services_downtimed.json")) as after:
                set_mocked_icinga_host_outputs(self.mocked_icinga_host, [before.read(), "", after.read(), ""])
        with self.icinga_hosts.services_downtimed("service.*", self.reason):
            assert_has_service_downtime_calls(
                self.mocked_icinga_host, [("host1", "service1"), ("host1", "service2")], self.reason
            )
            self.mocked_icinga_host.run_sync.reset_mock()

        self.mocked_icinga_host.run_sync.assert_called_with(
            'bash -c \'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1;service1" > '
            "/var/lib/icinga/rw/icinga.cmd '",
            'bash -c \'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1;service2" > '
            "/var/lib/icinga/rw/icinga.cmd '",
            print_output=False,
            print_progress_bars=False,
        )
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_services_downtime_is_kept_when_exception_is_raised(self, _mocked_time):
        """Downtime should not be removed if an exception is raised."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with pytest.raises(ValueError):
            with self.icinga_hosts.services_downtimed("service.*", self.reason):
                assert_has_service_downtime_calls(
                    self.mocked_icinga_host, [("host1", "service1"), ("host1", "service2")], self.reason
                )
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert not self.mocked_icinga_host.run_sync.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_services_downtime_is_removed_when_exception_is_raised(self, _mocked_time):
        """Downtime should not be removed if an exception is raised."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as before:
            with open(get_fixture_path("icinga", "status_with_services_downtimed.json")) as after:
                set_mocked_icinga_host_outputs(self.mocked_icinga_host, [before.read(), "", after.read(), ""])
        with pytest.raises(ValueError):
            with self.icinga_hosts.services_downtimed("service.*", self.reason, remove_on_error=True):
                assert_has_service_downtime_calls(
                    self.mocked_icinga_host, [("host1", "service1"), ("host1", "service2")], self.reason
                )
                self.mocked_icinga_host.run_sync.reset_mock()
                raise ValueError()
        assert self.mocked_icinga_host.run_sync.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_services_default_params(self, _mocked_time, caplog):
        """It should downtime the services on the Icinga server with the default params."""
        caplog.set_level(logging.INFO)
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.downtime_services(r"service\d", self.reason)
        assert_has_service_downtime_calls(
            self.mocked_icinga_host, [("host1", "service1"), ("host1", "service2")], self.reason
        )
        assert r'for services "service\d" for host: host1 (matched 2 unique service names on 1 host)' in caplog.text

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_downtime_services_custom_duration(self, _mocked_time):
        """It should downtime the services for the given duration on the Icinga server."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.downtime_services(r"service\d", self.reason, duration=timedelta(minutes=30))
        assert_has_service_downtime_calls(
            self.mocked_icinga_host, [("host1", "service1"), ("host1", "service2")], self.reason, duration=1800
        )

    def test_downtime_services_invalid_duration(self):
        """It should raise IcingaError if the duration is too short."""
        with pytest.raises(icinga.IcingaError, match="Downtime duration must be at least 1 minute"):
            self.icinga_hosts.downtime_services(r"service\d", self.reason, duration=timedelta(seconds=59))

    def test_downtime_services_unknown_host(self):
        """It should raise IcingaError if any host is not known to Icinga."""
        set_mocked_icinga_host_output(self.mocked_icinga_host, '{"host1": null}')
        with pytest.raises(
            icinga.IcingaError, match="Host host1 was not found in Icinga status - no hosts have been downtimed"
        ):
            self.icinga_hosts.downtime_services(r"service\d", self.reason)

    def test_downtime_services_unknown_service(self):
        """It should raise IcingaError if no services match the regex."""
        with open(get_fixture_path("icinga", "status_with_no_matching_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with pytest.raises(icinga.IcingaError, match=r'No services on host1 matched "service\\d"'):
            self.icinga_hosts.downtime_services(r"service\d", self.reason)

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_run_icinga_command(self, mocked_time):
        """It should run the specified command for all the hosts on the Icinga server."""
        self.icinga_hosts.run_icinga_command("TEST_COMMAND", "arg1", "arg2")
        call = "bash -c 'echo -n \"[1514764800] TEST_COMMAND;host1;arg1;arg2\" > /var/lib/icinga/rw/icinga.cmd '"

        self.mocked_icinga_host.run_sync.assert_called_once_with(call, print_output=False, print_progress_bars=False)
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_recheck_all_services(self, mocked_time):
        """It should force a recheck of all services for the hosts on the Icinga server."""
        self.icinga_hosts.recheck_all_services()
        self.mocked_icinga_host.run_sync.assert_called_once_with(
            'bash -c \'echo -n "[1514764800] SCHEDULE_FORCED_HOST_SVC_CHECKS;host1;1514764800" > '
            "/var/lib/icinga/rw/icinga.cmd '",
            print_output=False,
            print_progress_bars=False,
        )
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_recheck_failed_services_failed(self, mocked_time):
        """It should force a recheck of all services for the hosts on the Icinga server."""
        with open(get_fixture_path("icinga", "status_with_failed_services.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        self.icinga_hosts.recheck_failed_services()
        self.mocked_icinga_host.run_sync.assert_called_with(
            'bash -c \'echo -n "[1514764800] SCHEDULE_FORCED_SVC_CHECK;host2;check_name1;1514764800" > '
            "/var/lib/icinga/rw/icinga.cmd '",
            'bash -c \'echo -n "[1514764800] SCHEDULE_FORCED_SVC_CHECK;host2;check_name2;1514764800" > '
            "/var/lib/icinga/rw/icinga.cmd '",
            print_output=False,
            print_progress_bars=False,
        )
        assert mocked_time.called

    def test_recheck_failed_services_optimal(self):
        """It should force a recheck of all services for the hosts on the Icinga server."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        self.icinga_hosts.recheck_failed_services()
        # This also ensures that we are not making an additional call of run_sync in the recheck method
        self.mocked_icinga_host.run_sync.assert_called_with(
            Command('/usr/local/bin/icinga-status -j "host1"', ok_codes=[]),
            is_safe=True,
            print_output=False,
            print_progress_bars=False,
        )

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_remove_downtime(self, mocked_time):
        """It should remove the downtime for the hosts on the Icinga server."""
        self.icinga_hosts.remove_downtime()
        self.mocked_icinga_host.run_sync.assert_called_once_with(
            "bash -c 'echo -n \"[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1\" > /var/lib/icinga/rw/icinga.cmd '",
            print_output=False,
            print_progress_bars=False,
        )
        assert mocked_time.called

    @mock.patch("spicerack.icinga.time.time", return_value=1514764800)
    def test_remove_service_downtimes(self, mocked_time):
        """It should remove the downtime for the hosts on the Icinga server."""
        with open(get_fixture_path("icinga", "status_with_services_downtimed.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.remove_service_downtimes(r"service\d")
        self.mocked_icinga_host.run_sync.assert_called_with(
            *[
                f'bash -c \'echo -n "[1514764800] DEL_DOWNTIME_BY_HOST_NAME;host1;{service}" '
                f"> /var/lib/icinga/rw/icinga.cmd '"
                for service in ["service1", "service2"]
            ],
            print_output=False,
            print_progress_bars=False,
        )
        assert mocked_time.called

    def test_remove_service_downtimes_not_downtimed(self):
        """It should do nothing if the services exist but aren't downtimed."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        self.icinga_hosts.remove_service_downtimes(r"service\d")
        self.mocked_icinga_host.run_sync.assert_called_once()  # Only the icinga-status call, no downtimes.

    def test_remove_service_downtimes_unknown_service(self):
        """It should raise IcingaError if no services match the regex."""
        with open(get_fixture_path("icinga", "status_with_no_matching_services.json")) as f:
            set_mocked_icinga_host_outputs(self.mocked_icinga_host, [f.read(), "", "", ""])
        with pytest.raises(icinga.IcingaError, match=r'No services on host1 matched "service\\d"'):
            self.icinga_hosts.remove_service_downtimes(r"service\d")

    def test_get_status_ok(self):
        """It should parse the JSON payload and return an instance of HostsStatus."""
        with open(get_fixture_path("icinga", "status_with_failed_services.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        status = self.icinga_hosts.get_status()

        assert "--services" not in self.mocked_icinga_host.run_sync.call_args[0][0].command
        assert not status.optimal
        assert status.non_optimal_hosts == ["host2"]
        assert status.failed_services == {"host2": ["check_name1", "check_name2"]}
        assert status.failed_hosts == []

    def test_get_status_parse_fail(self):
        """It should raise IcingaStatusParseError if unable to parse the JSON payload."""
        with open(get_fixture_path("icinga", "status_invalid.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        with pytest.raises(icinga.IcingaStatusParseError, match="Unable to parse Icinga status"):
            self.icinga_hosts.get_status()

    def test_get_status_missing_hosts(self):
        """It should raise IcingaStatusNotFoundError if any host is missing its status."""
        with open(get_fixture_path("icinga", "status_missing.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())
        with pytest.raises(
            icinga.IcingaStatusNotFoundError,
            match="Host host2 was not found in Icinga status",
        ):
            self.icinga_hosts.get_status()

    def test_get_status_no_output(self):
        """It should raise IcingaError if there is no output from the icinga-status command."""
        self.mocked_icinga_host.run_sync.return_value = iter(())
        with pytest.raises(icinga.IcingaError, match="no output from icinga-status"):
            self.icinga_hosts.get_status()

    def test_get_status_with_services(self):
        """It should parse the JSON payload and return an instance of HostsStatus with service status."""
        with open(get_fixture_path("icinga", "status_with_services.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read())

        status = self.icinga_hosts.get_status(service_re=r"service\d")

        assert r"--services 'service\d'" in self.mocked_icinga_host.run_sync.call_args[0][0].command
        assert status.optimal
        assert {service["name"] for service in status["host1"].services} == {"service1", "service2"}
        assert status.failed_hosts == []

    def test_get_status_with_invalid_services_re(self):
        """It should raise re.error if the services regex is invalid."""
        with pytest.raises(re.error, match="nothing to repeat at position 0"):
            self.icinga_hosts.get_status(service_re="+")

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_optimal_ok(self, mocked_sleep):
        """It should return immediately if host is optimal."""
        with open(get_fixture_path("icinga", "status_valid.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read(), 2)

        self.icinga_hosts.wait_for_optimal()
        assert not mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_optimal_timeout(self, mocked_sleep):
        """It should raise icinga.IcingaError if host is optimal in the required time."""
        with open(get_fixture_path("icinga", "status_with_failed_services.json")) as f:
            set_mocked_icinga_host_output(self.mocked_icinga_host, f.read(), 20)

        with pytest.raises(icinga.IcingaError, match="Not all services are recovered"):
            self.icinga_hosts.wait_for_optimal()

        assert mocked_sleep.called


def _get_hoststatus(hostname, down=False, failed=False):
    """Return an instance of HostStatus suitable for testing based on the parameters.

    Arguments:
        hostname: the name to use for the host.
        down: whether the reported state should be UP or DOWN.
        failed: whether if should have some failed services or not.

    """
    params = {
        "name": hostname,
        "state": "UP",
        "optimal": True,
        "failed_services": [],
        "downtimed": False,
        "notifications_enabled": True,
    }

    failed_services = [
        {"host": hostname, "name": "check_name1", "status": {}},
        {"host": hostname, "name": "check_name2", "status": {}},
    ]

    if down:
        params["state"] = "DOWN"
        params["optimal"] = False

    if failed:
        params["failed_services"] = failed_services
        params["optimal"] = False

    return icinga.HostStatus(**params)


def test_hoststatus_failed_services():
    """It should return the list of check names that are failed."""
    status = _get_hoststatus("hostname", failed=True)
    assert status.failed_services == ["check_name1", "check_name2"]


class TestHostsStatus:
    """Tests for the HostsStatus class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.status = {
            "ok": icinga.HostsStatus(),  # All optimal hosts
            "down": icinga.HostsStatus(),  # Some hosts are down, no hosts have failed services
            "failed": icinga.HostsStatus(),  # All hosts up, but some have failed services
            "down_failed": icinga.HostsStatus(),  # Some hosts are donw, some have failed services
        }

        for i in range(1, 4):
            hostname = f"host{i}"
            self.status["ok"][hostname] = _get_hoststatus(hostname)
            self.status["down"][hostname] = _get_hoststatus(hostname)
            self.status["failed"][hostname] = _get_hoststatus(hostname)
            self.status["down_failed"][hostname] = _get_hoststatus(hostname)

        for i in range(4, 6):
            hostname = f"host{i}"
            self.status["down"][hostname] = _get_hoststatus(hostname, down=True)
            self.status["down_failed"][hostname] = _get_hoststatus(hostname, down=True)

        for i in range(6, 8):
            hostname = f"host{i}"
            self.status["failed"][hostname] = _get_hoststatus(hostname, failed=True)
            self.status["down_failed"][hostname] = _get_hoststatus(hostname, failed=True)

    def test_optimal_ok(self):
        """It should return True if all hosts are optimal."""
        assert self.status["ok"].optimal

    @pytest.mark.parametrize("key", ("down", "failed", "down_failed"))
    def test_optimal_ko(self, key):
        """If should return False if any host is not optimal."""
        assert not self.status[key].optimal

    @pytest.mark.parametrize(
        "key, expected",
        (
            ("ok", []),
            ("down", ["host4", "host5"]),
            ("failed", ["host6", "host7"]),
            ("down_failed", ["host4", "host5", "host6", "host7"]),
        ),
    )
    def test_non_optimal_hosts(self, key, expected):
        """It should return the list of hostnames with non optimal status."""
        assert sorted(self.status[key].non_optimal_hosts) == expected

    @pytest.mark.parametrize(
        "key, expected",
        (
            ("ok", {}),
            ("down", {"host4": [], "host5": []}),
            (
                "failed",
                {
                    "host6": ["check_name1", "check_name2"],
                    "host7": ["check_name1", "check_name2"],
                },
            ),
            (
                "down_failed",
                {
                    "host4": [],
                    "host5": [],
                    "host6": ["check_name1", "check_name2"],
                    "host7": ["check_name1", "check_name2"],
                },
            ),
        ),
    )
    def test_failed_services(self, key, expected):
        """It should return the dictionary with the hostnames as key and the list of failed services as values."""
        assert self.status[key].failed_services == expected

    @pytest.mark.parametrize(
        "key, expected",
        (
            ("ok", []),
            ("down", ["host4", "host5"]),
            ("failed", []),
            ("down_failed", ["host4", "host5"]),
        ),
    )
    def test_failed_hosts(self, key, expected):
        """It should return the list of hostnames with non optimal status."""
        assert sorted(self.status[key].failed_hosts) == expected



"""Log module."""
import logging
from pathlib import Path
from typing import Optional

from wmflib.irc import SALSocketHandler

root_logger = logging.getLogger()
irc_logger = logging.getLogger("spicerack_irc_announce")


class FilterOutCumin(logging.Filter):
    """A logging output filter to filter out Cumin's logs."""

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter out Cumin's log messages.

        Arguments:
            record (logging.LogRecord):

        Returns:
            int: :py:data:`False` if the record should be filtered out, :py:data:`True` if it should be included.
            According to Python's logging interface, see: https://docs.python.org/3/library/logging.html#filter-objects

        """
        if record.name == "cumin" or record.name.startswith("cumin."):
            return False  # Filter it out

        return True


def setup_logging(
    base_path: Path,
    name: str,
    user: str,
    dry_run: bool = True,
    host: Optional[str] = None,
    port: int = 0,
) -> None:
    """Setup the root logger instance.

    Arguments:
        base_path (str, pathlib.Path): the base path where to save the logs.
        name (str): the name of log file to use without extension.
        user (str): the username for the IRC logging.
        dry_run (bool, optional): whether this is a dry-run.
        host (str, optional): the tcpircbot hostname for the IRC logging.
        port (int, optional): the tcpircbot port for the IRC logging.

    """
    logging.raiseExceptions = False
    base_path.mkdir(mode=0o755, parents=True, exist_ok=True)

    if dry_run:
        dry_run_prefix = "DRY-RUN "
    else:
        dry_run_prefix = ""

    # Default INFO logging
    formatter = logging.Formatter(fmt=f"%(asctime)s {dry_run_prefix}{user} %(process)d [%(levelname)s] %(message)s")
    handler = logging.FileHandler(base_path / f"{name}.log")
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)

    # Extended logging for detailed debugging
    formatter_extended = logging.Formatter(
        fmt=(
            f"%(asctime)s {dry_run_prefix}{user} %(process)d [%(levelname)s %(filename)s:%(lineno)s in %(funcName)s] "
            "%(message)s"
        )
    )
    handler_extended = logging.FileHandler(base_path / f"{name}-extended.log")
    handler_extended.setFormatter(formatter_extended)
    handler_extended.setLevel(logging.DEBUG)

    # Stderr logging
    output_handler = logging.StreamHandler()
    if dry_run:
        output_handler.setFormatter(logging.Formatter(fmt="DRY-RUN: %(message)s"))
        output_handler.setLevel(logging.DEBUG)
    else:
        output_handler.setLevel(logging.INFO)
    output_handler.addFilter(FilterOutCumin())

    root_logger.addHandler(handler)
    root_logger.addHandler(handler_extended)
    root_logger.addHandler(output_handler)
    root_logger.setLevel(logging.DEBUG)

    if not dry_run and host is not None and port > 0:
        irc_logger.addHandler(SALSocketHandler(host, port, user))
        irc_logger.setLevel(logging.INFO)

    # Silence external noisy loggers
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)

    # Elasticsearch lib is very noisy about HTTP level errors
    # ideally, we'd want to keep it at WARNING level for logs
    # sent to file, but ERROR for the console. Since this is
    # non trivial, let's raise level to ERROR for the moment.
    logging.getLogger("elasticsearch").setLevel(logging.ERROR)


def log_task_start(message: str) -> None:
    """Log the start of a task both on the logs and IRC.

    Arguments:
        message (str): the message to be logged.

    """
    irc_logger.info("START - %s", message)


def log_task_end(status: str, message: str) -> None:
    """Log the start of a task both on the logs and IRC.

    Arguments:
        status (str): the final status of the task.
        message (str): the message to be logged.

    """
    irc_logger.info("END (%s) - %s", status, message)



"""Custom Spicerack type hints."""
from typing import Sequence, TypeVar

from cumin import NodeSet

TypeHosts = TypeVar("TypeHosts", Sequence[str], NodeSet)
""":py:class:`typing.TypeVar` for hosts, can be a :py:class:`ClusterShell.NodeSet.NodeSet` or any sequence of
:py:class:`str`."""



"""Initialization tests."""
import logging
from unittest import mock

import pytest
from git import Repo
from requests import Session
from wmflib.actions import ActionsDict
from wmflib.dns import Dns
from wmflib.phabricator import Phabricator
from wmflib.prometheus import Prometheus

from spicerack import Spicerack
from spicerack.administrative import Reason
from spicerack.alerting import AlertingHosts
from spicerack.alertmanager import AlertmanagerHosts
from spicerack.confctl import ConftoolEntity
from spicerack.debmonitor import Debmonitor
from spicerack.dhcp import DHCP
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import ElasticsearchClusters
from spicerack.exceptions import SpicerackError
from spicerack.ganeti import Ganeti
from spicerack.icinga import IcingaHosts
from spicerack.ipmi import Ipmi
from spicerack.kafka import Kafka
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.mysql_legacy import MysqlLegacy
from spicerack.netbox import Netbox, NetboxServer
from spicerack.puppet import PuppetHosts, PuppetMaster
from spicerack.redfish import RedfishDell
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote, RemoteHosts
from spicerack.reposync import RepoSync
from spicerack.tests import SPICERACK_TEST_PARAMS, get_fixture_path
from spicerack.toolforge.etcdctl import EtcdctlController


@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("wmflib.dns.resolver", autospec=True)
def test_spicerack(mocked_dns_resolver, mocked_remote_query, monkeypatch):
    """An instance of Spicerack should allow to access all the library features."""
    monkeypatch.setenv("SUDO_USER", "user1")
    verbose = True
    dry_run = False
    proxy = "http://proxy.example.com:8080"
    spicerack = Spicerack(verbose=verbose, dry_run=dry_run, http_proxy=proxy, **SPICERACK_TEST_PARAMS)

    assert spicerack.verbose is verbose
    assert spicerack.dry_run is dry_run
    assert spicerack.username == "user1"
    assert spicerack.config_dir == get_fixture_path()
    assert spicerack.http_proxy == proxy
    assert spicerack.requests_proxies == {"http": proxy, "https": proxy}
    assert isinstance(spicerack.irc_logger, logging.Logger)
    assert isinstance(spicerack.actions, ActionsDict)
    assert isinstance(spicerack.remote(), Remote)
    assert isinstance(spicerack.remote(installer=True), Remote)
    assert isinstance(spicerack.confctl("discovery"), ConftoolEntity)
    assert isinstance(spicerack.confctl("mwconfig"), ConftoolEntity)
    assert isinstance(spicerack.dns(), Dns)
    assert isinstance(spicerack.discovery("discovery-record"), Discovery)
    assert isinstance(spicerack.mediawiki(), MediaWiki)
    assert isinstance(spicerack.mysql(), Mysql)
    assert isinstance(spicerack.mysql_legacy(), MysqlLegacy)
    assert isinstance(spicerack.redis_cluster("cluster"), RedisCluster)
    assert isinstance(
        spicerack.elasticsearch_clusters("search_eqiad", ("some_core_dc",)),
        ElasticsearchClusters,
    )
    assert isinstance(spicerack.admin_reason("Reason message", task_id="T12345"), Reason)
    assert isinstance(spicerack.puppet(mock.MagicMock(spec_set=RemoteHosts)), PuppetHosts)
    assert isinstance(
        spicerack.phabricator(get_fixture_path("phabricator", "valid.conf")),
        Phabricator,
    )
    assert isinstance(spicerack.prometheus(), Prometheus)
    assert isinstance(spicerack.debmonitor(), Debmonitor)
    assert isinstance(spicerack.ganeti(), Ganeti)
    assert isinstance(spicerack.requests_session("name"), Session)
    assert isinstance(
        spicerack.etcdctl(remote_host=mock.MagicMock(spec_set=RemoteHosts)),
        EtcdctlController,
    )
    assert isinstance(spicerack.kafka(), Kafka)
    assert isinstance(spicerack.alertmanager_hosts(["host1", "host2"]), AlertmanagerHosts)
    assert mocked_remote_query.called
    assert mocked_dns_resolver.Resolver.called


def test_spicerack_http_proxy():
    """An instance of Spicerack by default should not set any HTTP proxy."""
    spicerack = Spicerack(**SPICERACK_TEST_PARAMS)
    assert spicerack.http_proxy == ""
    assert spicerack.requests_proxies is None


@mock.patch("spicerack.gethostname", return_value="test.example.com")
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("spicerack.icinga.CommandFile", autospec=True)
def test_spicerack_icinga(mocked_command_file, mocked_remote_query, mocked_dns, mocked_hostname, monkeypatch):
    """An instance of Spicerack should allow to get an Icinga and IcingaHosts instances."""
    monkeypatch.setenv("SUDO_USER", "user1")
    icinga_server = mock.MagicMock(spec_set=RemoteHosts)
    icinga_server.hosts = "icinga-server.example.com"
    icinga_server.__len__.return_value = 1
    mocked_remote_query.return_value = icinga_server
    mocked_dns.return_value.resolve_cname.return_value = "icinga-server.example.com"
    mocked_command_file.return_value = "/var/lib/icinga/rw/icinga.cmd"

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert spicerack.icinga_master_host.hosts == "icinga-server.example.com"
    assert isinstance(spicerack.icinga_hosts(["host1", "host2"]), IcingaHosts)
    mocked_hostname.assert_called_once_with()


@mock.patch("spicerack.get_puppet_ca_hostname", return_value="puppetmaster.example.com")
@mock.patch("spicerack.remote.Remote.query", autospec=True)
def test_spicerack_puppet_master(mocked_remote_query, mocked_get_puppet_ca_hostname):
    """An instance of Spicerack should allow to get a PuppetMaster instance."""
    host = mock.MagicMock(spec_set=RemoteHosts)
    host.__len__.return_value = 1
    mocked_remote_query.return_value = host
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert isinstance(spicerack.puppet_master(), PuppetMaster)
    mocked_get_puppet_ca_hostname.assert_called_once_with()
    assert mocked_remote_query.called


def test_spicerack_management_password_instances(monkeypatch):
    """Should instantiate the instances that require the management password."""
    monkeypatch.setenv("MGMT_PASSWORD", "env_password")
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert spicerack.management_password == "env_password"
    assert isinstance(spicerack.ipmi("test-mgmt.example.com"), Ipmi)
    assert isinstance(spicerack.redfish("test-mgmt.example.com", "root"), RedfishDell)
    assert isinstance(spicerack.redfish("test-mgmt.example.com", "root", "other_password"), RedfishDell)


def test_spicerack_management_password_cached(monkeypatch):
    """Should ask for the management_password only once and cache its result."""
    monkeypatch.setenv("MGMT_PASSWORD", "first_password")
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert spicerack.management_password == "first_password"
    monkeypatch.setenv("MGMT_PASSWORD", "second_password")
    assert spicerack.management_password == "first_password"


@pytest.mark.parametrize(
    "read_write, token",
    (
        (False, "ro_token"),
        (True, "rw_token"),
    ),
)
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("pynetbox.api")
def test_spicerack_netbox(mocked_pynetbox, mocked_remote_query, mocked_dns, read_write, token):
    """Test instantiating Netbox abstraction."""
    netbox_server = mock.MagicMock(spec_set=RemoteHosts)
    netbox_server.hosts = "netbox-server.example.com"
    mocked_remote_query.return_value = netbox_server
    mocked_dns.return_value.resolve_cname.return_value = "netbox-server.example.com"
    mocked_pynetbox.reset_mock()

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert isinstance(spicerack.netbox(read_write=read_write), Netbox)
    # Values from fixtures/netbox/config.yaml
    mocked_pynetbox.assert_called_once_with("https://netbox.example.com", token=token)
    assert spicerack.netbox_master_host.hosts == "netbox-server.example.com"

    mocked_pynetbox.reset_mock()
    mocked_pynetbox.return_value.dcim.devices.get.return_value.device_role.slug = "server"
    assert isinstance(spicerack.netbox_server("host1"), NetboxServer)


def test_spicerack_dhcp():
    """Test spicerack.dhcp. It should succeed if a host list with more than one member is passed."""
    mock_hosts = mock.MagicMock()
    mock_hosts.__len__.return_value = 1

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert isinstance(spicerack.dhcp(mock_hosts), DHCP)


def test_run_cookbook_no_callback():
    """It should raise a SpicerackError if there is no get_cookbook_callback defined."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    with pytest.raises(SpicerackError, match="Unable to run other cookbooks, get_cookbook_callback is not set."):
        spicerack.run_cookbook("class_api.example", [])


@mock.patch("spicerack.Path.is_dir")
@mock.patch("spicerack.Repo")
def test_reposync(mocked_repo, mocked_is_dir):
    """Test spicerack.reposync."""
    spicerack = Spicerack(**SPICERACK_TEST_PARAMS)

    repo = mock.MagicMock(spec_set=Repo)
    repo.bare = False

    with pytest.raises(SpicerackError, match="Unknown repo missing_repo"):
        spicerack.reposync("missing_repo")

    mocked_is_dir.return_value = False
    with pytest.raises(SpicerackError, match=r"The repo directory \(/testrepo\) does not exist"):
        spicerack.reposync("testrepo")

    mocked_is_dir.return_value = True
    mocked_repo.return_value = repo
    with pytest.raises(SpicerackError, match=r"The repo directory \(/testrepo\) is not a bare git repository"):
        spicerack.reposync("testrepo")

    repo.bare = True
    assert isinstance(spicerack.reposync("testrepo"), RepoSync)


@mock.patch("spicerack.gethostname", return_value="test.example.com")
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("spicerack.icinga.CommandFile", autospec=True)
def test_spicerack_alerting(mocked_command_file, mocked_remote_query, mocked_dns, mocked_hostname, monkeypatch):
    """An instance of Spicerack should allow to get an AlertingHosts instance."""
    monkeypatch.setenv("SUDO_USER", "user1")
    icinga_server = mock.MagicMock(spec_set=RemoteHosts)
    icinga_server.hosts = "icinga-server.example.com"
    icinga_server.__len__.return_value = 1
    mocked_remote_query.return_value = icinga_server
    mocked_dns.return_value.resolve_cname.return_value = "icinga-server.example.com"
    mocked_command_file.return_value = "/var/lib/icinga/rw/icinga.cmd"

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert spicerack.icinga_master_host.hosts == "icinga-server.example.com"
    assert isinstance(spicerack.alerting_hosts(["host1", "host2"]), AlertingHosts)
    mocked_hostname.assert_called_once_with()



"""Group3 argparse."""
import argparse

__title__ = __doc__


def argument_parser():
    """As required by spicerack._cookbook."""
    return argparse.ArgumentParser("group3.argparse")


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""



"""MySQL module (legacy).

Todo:
    replace with a proper MySQL module that uses a Python MySQL client, preferably in a parallel way.

"""
import logging
from datetime import datetime
from typing import Dict, Iterator, Optional, Tuple, Union

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet

from spicerack.constants import CORE_DATACENTERS
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError
from spicerack.remote import Remote, RemoteHostsAdapter

REPLICATION_ROLES: Tuple[str, ...] = ("master", "slave", "standalone")
"""tuple: list of valid replication roles."""
CORE_SECTIONS: Tuple[str, ...] = (
    "s1",
    "s2",
    "s3",
    "s4",
    "s5",
    "s6",
    "s7",
    "s8",
    "x1",
    "x2",
    "es4",
    "es5",
)
"""tuple: list of valid MySQL RW section names (external storage RO sections are not included here)."""
ACTIVE_ACTIVE_SECTIONS: Tuple[str] = ("x2",)
"""tuple: active/active sections that should not go read-only"""

logger = logging.getLogger(__name__)


class MysqlLegacyError(SpicerackError):
    """Custom exception class for errors of this module."""


class MysqlLegacyRemoteHosts(RemoteHostsAdapter):
    """Custom RemoteHosts class to execute MySQL queries."""

    def run_query(  # pylint: disable=too-many-arguments
        self,
        query: str,
        database: str = "",
        success_threshold: float = 1.0,
        batch_size: Optional[Union[int, str]] = None,
        batch_sleep: Optional[float] = None,
        is_safe: bool = False,
    ) -> Iterator[Tuple[NodeSet, MsgTreeElem]]:
        """Execute the query via Remote.

        Arguments:
            query (str): the mysql query to be executed. Double quotes must be already escaped.
            database (str, optional): an optional MySQL database to connect to before executing the query.
            success_threshold (float, optional): to consider the execution successful, must be between 0.0 and 1.0.
            batch_size (int, str, optional): the batch size for cumin, either as percentage (e.g. ``25%``) or absolute
                number (e.g. ``5``).
            batch_sleep (float, optional): the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe (bool, optional): whether the command is safe to run also in dry-run mode because it's a read-only
                command that doesn't modify the state.

        Returns:
            generator: as returned by :py:meth:`cumin.transports.BaseWorker.get_results`.

        Raises:
            RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        command = f'mysql --skip-ssl --skip-column-names --batch -e "{query}" {database}'.strip()
        return self._remote_hosts.run_sync(
            command,
            success_threshold=success_threshold,
            batch_size=batch_size,
            batch_sleep=batch_sleep,
            is_safe=is_safe,
        )


class MysqlLegacy:
    """Class to manage MySQL servers."""

    heartbeat_query = (
        "SELECT ts FROM heartbeat.heartbeat WHERE datacenter = '{dc}' and shard = '{section}' "
        "ORDER BY ts DESC LIMIT 1"
    )
    """str: Query pattern to check the heartbeat for a given datacenter and section."""

    def __init__(self, remote: Remote, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            remote (spicerack.remote.Remote): the Remote instance.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._remote = remote
        self._dry_run = dry_run

    def get_dbs(self, query: str) -> MysqlLegacyRemoteHosts:
        """Get a MysqlLegacyRemoteHosts instance for the matching hosts.

        Arguments:
            query (str): the Remote query to use to fetch the DB hosts.

        Returns:
            spicerack.mysql_legacy.MysqlLegacyRemoteHosts: an instance with the remote targets.

        """
        return MysqlLegacyRemoteHosts(self._remote.query(query))

    def get_core_dbs(
        self,
        *,
        datacenter: Optional[str] = None,
        section: Optional[str] = None,
        replication_role: Optional[str] = None,
        excludes: Tuple[str, ...] = (),
    ) -> MysqlLegacyRemoteHosts:
        """Find the core databases matching the parameters.

        Arguments:
            datacenter (str, optional): the name of the datacenter to filter for, accepted values are those specified in
                :py:data:`spicerack.constants.CORE_DATACENTERS`.
            replication_role (str, optional): the repication role to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql_legacy.REPLICATION_ROLES`.
            section (str, optional): a specific section to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql_legacy.CORE_SECTIONS`.
            excludes (Tuple[str, ...]): sections to exclude from getting

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on invalid data or unexpected matching hosts.

        Returns:
            spicerack.mysql_legacy.MysqlLegacyRemoteHosts: an instance with the remote targets.

        """
        query_parts = ["A:db-core"]
        dc_multipler = len(CORE_DATACENTERS)
        section_multiplier = len(CORE_SECTIONS)

        if datacenter is not None:
            dc_multipler = 1
            if datacenter not in CORE_DATACENTERS:
                raise MysqlLegacyError(f"Got invalid datacenter {datacenter}, accepted values are: {CORE_DATACENTERS}")

            query_parts.append("A:" + datacenter)

        for exclude in excludes:
            if exclude not in CORE_SECTIONS:
                raise MysqlLegacyError(f"Got invalid excludes {exclude}, accepted values are: {CORE_SECTIONS}")
            section_multiplier -= 1
            query_parts.append(f"not A:db-section-{exclude}")

        if section is not None:
            section_multiplier = 1
            if section not in CORE_SECTIONS:
                raise MysqlLegacyError(f"Got invalid section {section}, accepted values are: {CORE_SECTIONS}")

            query_parts.append(f"A:db-section-{section}")

        if replication_role is not None:
            if replication_role not in REPLICATION_ROLES:
                raise MysqlLegacyError(
                    f"Got invalid replication_role {replication_role}, accepted values are: {REPLICATION_ROLES}"
                )

            query_parts.append(f"A:db-role-{replication_role}")

        mysql_hosts = MysqlLegacyRemoteHosts(self._remote.query(" and ".join(query_parts)))

        # Sanity check of matched hosts in case of master selection
        if replication_role == "master" and len(mysql_hosts) != dc_multipler * section_multiplier:
            raise MysqlLegacyError(f"Matched {len(mysql_hosts)} masters, expected {dc_multipler * section_multiplier}")

        return mysql_hosts

    def set_core_masters_readonly(self, datacenter: str) -> None:
        """Set the core masters in read-only.

        Arguments:
            datacenter (str): the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql_legacy.MysqlLegacyError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-only", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master", excludes=ACTIVE_ACTIVE_SECTIONS)
        target.run_query("SET GLOBAL read_only=1")
        self.verify_core_masters_readonly(datacenter, True)

    def set_core_masters_readwrite(self, datacenter: str) -> None:
        """Set the core masters in read-write.

        Arguments:
            datacenter (str): the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql_legacy.MysqlLegacyError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-write", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master", excludes=ACTIVE_ACTIVE_SECTIONS)
        target.run_query("SET GLOBAL read_only=0")
        self.verify_core_masters_readonly(datacenter, False)

    def verify_core_masters_readonly(self, datacenter: str, is_read_only: bool) -> None:
        """Verify that the core masters are in read-only or read-write mode.

        Arguments:
            datacenter (str): the name of the datacenter to filter for.
            is_read_only (bool): whether the read-only mode should be set or not.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure.

        """
        logger.debug(
            "Verifying core DB masters in %s have read-only=%d",
            datacenter,
            is_read_only,
        )
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master", excludes=ACTIVE_ACTIVE_SECTIONS)
        expected = str(int(is_read_only))  # Convert it to the returned value from MySQL: 1 or 0.
        failed = False

        for nodeset, output in target.run_query("SELECT @@global.read_only", is_safe=True):
            response = output.message().decode()
            if response != expected:
                logger.error(
                    "Expected output to be '%s', got '%s' for hosts %s",
                    expected,
                    response,
                    str(nodeset),
                )
                failed = True

        if failed and not self._dry_run:
            raise MysqlLegacyError(
                f"Verification failed that core DB masters in {datacenter} have read-only={is_read_only}"
            )

    def check_core_masters_in_sync(self, dc_from: str, dc_to: str) -> None:
        """Check that all core masters in dc_to are in sync with the core masters in dc_from.

        Arguments:
            dc_from (str): the name of the datacenter from where to get the master positions.
            dc_to (str): the name of the datacenter where to check that they are in sync.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        logger.debug("Waiting for the core DB masters in %s to catch up", dc_to)
        heartbeats = self.get_core_masters_heartbeats(dc_from, dc_from)
        self.check_core_masters_heartbeats(dc_to, dc_from, heartbeats)

    def get_core_masters_heartbeats(self, datacenter: str, heartbeat_dc: str) -> Dict[str, datetime]:
        """Get the current heartbeat values from core DB masters in DC for a given heartbeat DC.

        Arguments:
            datacenter (str): the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc (str): the name of the datacenter for which to filter the heartbeat query.

        Returns:
            dict: a dictionary with the section name :py:class:`str` as keys and their heartbeat
            :py:class:`datetime.datetime` as values. For example::

                {'s1': datetime.datetime(2018, 1, 2, 11, 22, 33, 123456)}

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        heartbeats = {}
        for section in CORE_SECTIONS:
            core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
            heartbeats[section] = MysqlLegacy._get_heartbeat(core_dbs, section, heartbeat_dc)

        return heartbeats

    def check_core_masters_heartbeats(
        self, datacenter: str, heartbeat_dc: str, heartbeats: Dict[str, datetime]
    ) -> None:
        """Check the current heartbeat values in the core DB masters in DC are in sync with the provided heartbeats.

        Arguments:
            datacenter (str): the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc (str): the name of the datacenter for which to filter the heartbeat query.
            heartbeats (dict): a dictionary with the section name :py:class:`str` as keys and heartbeat
                :py:class:`datetime.datetime` for each core section as values.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        for section, heartbeat in heartbeats.items():
            self._check_core_master_in_sync(datacenter, heartbeat_dc, section, heartbeat)

    @retry(exceptions=(MysqlLegacyError,))
    def _check_core_master_in_sync(
        self,
        datacenter: str,
        heartbeat_dc: str,
        section: str,
        parent_heartbeat: datetime,
    ) -> None:
        """Check and retry that the heartbeat value in a core DB master in DC is in sync with the provided heartbeat.

        Arguments:
            datacenter (str): the name of the datacenter from where to get the heartbeat value.
            heartbeat_dc (str): the name of the datacenter for which to filter the heartbeat query.
            section (str): the section name from where to get the heartbeat value and filter the heartbeat query.
            master_heartbeat (datetime.datetime): the reference heartbeat from the parent master to use to verify this
                master is in sync with it.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime
                or not yet in sync.

        """
        core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
        local_heartbeat = MysqlLegacy._get_heartbeat(core_dbs, section, heartbeat_dc)

        # The check requires that local_heartbeat is stricly greater than parent_heartbeat because heartbeat writes also
        # when the DB is in read-only mode and has a granularity of 1s (as of 2018-09), meaning that an event could have
        # been written after the last heartbeat but before the DB was set in read-only mode and that event could not
        # have been replicated, hence checking the next heartbeat to ensure they are in sync.
        if local_heartbeat <= parent_heartbeat:
            delta = (local_heartbeat - parent_heartbeat).total_seconds()
            raise MysqlLegacyError(
                f"Heartbeat from master {core_dbs} for section {section} not yet in sync: "
                f"{local_heartbeat} <= {parent_heartbeat} (delta={delta})"
            )

    @staticmethod
    def _get_heartbeat(mysql_hosts: MysqlLegacyRemoteHosts, section: str, heartbeat_dc: str) -> datetime:
        """Get the heartbeat from the remote host for a given DC.

        Arguments:
            mysql_hosts (spicerack.mysql_legacy.MysqlLegacyRemoteHosts): the instance for the target DB to query.
            section (str): the DB section for which to get the heartbeat.
            heartbeat_dc (str): the name of the datacenter for which to filter the heartbeat query.

        Returns:
            datetime.datetime: the converted heartbeat.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        query = MysqlLegacy.heartbeat_query.format(dc=heartbeat_dc, section=section)

        for _, output in mysql_hosts.run_query(query, is_safe=True):
            try:
                heartbeat_str = output.message().decode()
                heartbeat = datetime.strptime(heartbeat_str, "%Y-%m-%dT%H:%M:%S.%f")
                break
            except (TypeError, ValueError) as e:
                raise MysqlLegacyError(f"Unable to convert heartbeat '{heartbeat_str}' into datetime") from e
        else:
            raise MysqlLegacyError(f"Unable to get heartbeat from master {mysql_hosts} for section {section}")

        return heartbeat



"""Manage updates to automated git repositories."""
import os
from contextlib import contextmanager
from distutils.dir_util import copy_tree
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator, Optional

from git import Actor, Repo
from git.exc import GitError
from git.remote import PushInfo
from wmflib.interactive import ask_confirmation

from spicerack.constants import KEYHOLDER_SOCK
from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteHosts

logger = getLogger(__name__)


class RepoSyncError(SpicerackError):
    """Exception raised for push errors."""


class RepoSyncPushError(RepoSyncError):
    """Exception raised for push errors."""


class RepoSyncNoChangeError(RepoSyncError):
    """Exception raised for push errors."""


class RepoSync:
    """Class for syncing git repos."""

    _data_subdir: str = "data"
    _email: str = "noc@wikimedia.org"

    def __init__(self, repo: Repo, username: str, remote_hosts: RemoteHosts, *, dry_run: bool) -> None:
        """Initialise the object.

        Arguments:
            repo (git.repo.base.Repo): the path to the repo
            username (str): The username making the change
            remote_hosts (spicerack.remote.RemoteHosts): A remotes hosts object
                pointing to all git remote servers.  The servers are expected to
                have a valid KEYHOLDER_SOCK configured
            dry_run (bool): don't preform any write actions

        """
        self._hexsha: Optional[str] = None
        self._username = username
        self._remote_hosts = remote_hosts
        self._dry_run = dry_run
        self._repo = repo
        self._author = Actor(self._username, self._email)

    @property
    def hexsha(self) -> Optional[str]:
        """Returns the hexsha of the last commit."""
        return self._hexsha

    def _commit(self, working_repo: Repo, message: str) -> None:
        """Commit files in working repo.

        Arguments:
            working_repo (git.repo.base.Repo): The working git repo
            message (str): the commit message

        Raises:
            spicerack.reposync.RepoSyncNoChangeError: if no changes detected
            spicerack.reposync.RepoSyncError: If no hexsha returned for commit

        """
        # Don't use working_repo.index.add(".") as it adds the .git folder
        # https://github.com/gitpython-developers/GitPython/issues/292
        working_repo.git.add(A=True)
        if working_repo.head.is_valid() and not working_repo.index.diff(working_repo.head.commit):
            raise RepoSyncNoChangeError("Nothing to commit")
        commit = working_repo.index.commit(message, author=self._author, committer=self._author)
        if not isinstance(commit.hexsha, str):
            raise RepoSyncError("No valid commit hexsha from commit")
        logger.info("Committed changes: %s", commit.hexsha)
        self._hexsha = commit.hexsha

    def _push(self, working_repo: Repo) -> None:
        """Push the committed changes to the repository's remote.

        Arguments:
            working_repo (git.repo.base.Repo): the repository with the commit to push.

        Raises:
            spicerack.reposync.RepoSyncPushError: if there was an error pushing

        """
        if self._dry_run:
            logger.info("Would have pushed commit")
            return

        old_ssh_auth_sock = os.getenv("SSH_AUTH_SOCK")
        os.environ["SSH_AUTH_SOCK"] = KEYHOLDER_SOCK

        try:
            for remote in working_repo.remotes:
                logger.debug("Attempt push to: %s", remote)
                try:
                    # TODO: later versions of git python have raise_if_error
                    push_info_list = remote.push()
                    for push_info in push_info_list:
                        msg = f"bitflags {push_info.flags}: {push_info.summary.strip()}"
                        if push_info.flags & PushInfo.ERROR == PushInfo.ERROR:
                            raise RepoSyncPushError(f"Error pushing to {remote}: {msg}")
                # remote.push returns an empty list on error
                except (StopIteration, GitError) as error:
                    raise RepoSyncPushError(f"Error pushing to {remote}: {error}") from error
                logger.info("Pushed to %s", remote)
        finally:
            if old_ssh_auth_sock is None:
                del os.environ["SSH_AUTH_SOCK"]
            else:
                os.environ["SSH_AUTH_SOCK"] = old_ssh_auth_sock

    def _update_local(self, working_dir: Path, message: str) -> None:
        """Update the repo with data from fetch_data.

        Arguments:
            working_dir (pathlib.Path): The temporary directory used to build diffs
            message (str): the commit message

        """
        repo_dir = working_dir / "repo"
        data_dir = working_dir / self._data_subdir
        working_repo = self._repo.clone(repo_dir)
        # Delete all existing files to ensure removal of stale data
        working_repo.git.rm("./", r=True, ignore_unmatch=True)
        # TODO: on python >= 3.8 we can use shutil.copytree with dirs_exist_ok=True
        copy_tree(str(data_dir), str(repo_dir), preserve_symlinks=1)
        self._commit(working_repo, message)
        print(working_repo.git.show(["--color=always", "HEAD"]))
        ask_confirmation(f"Ok to push changes to {self._repo.common_dir}")
        self._push(working_repo)

    @contextmanager
    def update(self, message: str) -> Generator:
        """Context manager for updating a temporary directory with new data.

        The context manager will create and yield a temporary directory.  Users should populate
        this directory with fresh data to be committed to the main repository.

        Arguments:
            message (str): the commit message

        Yields:
            pathlib.Path: temporary directory to populate with data intended for the git repo

        Raises:
            spicerack.reposync.RepoSyncError: if unable to update local repo

        """
        with TemporaryDirectory() as tmp_dir:
            working_dir = Path(tmp_dir)
            data_dir = working_dir / self._data_subdir
            data_dir.mkdir()
            yield data_dir
            try:
                next(data_dir.iterdir())
            except StopIteration:
                raise RepoSyncError("No data written to data directory") from None
            try:
                self._update_local(working_dir, message)
            except RepoSyncNoChangeError as error:
                logger.info(error)
                return
        logger.debug("Push to remotes: %s", self._repo.remotes)
        self._push(self._repo)



"""Provide an interface for manipulating DHCP configuration snippets for our dynamic/temporary DHCP system."""

import base64
import re
import textwrap
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass
from hashlib import sha256
from ipaddress import IPv4Address
from typing import Iterator

from wmflib.constants import ALL_DATACENTERS

from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteExecutionError, RemoteHosts

DHCP_TARGET_PATH = "/etc/dhcp/automation"
"""str: The path to the top of the DHCPd automation directory."""


MGMT_HOSTNAME_RE = r"\.mgmt\.{dc}\.wmnet"
"""str: A regular expression when formatted with a `dc` parameter will match a management hostname."""


class DHCPError(SpicerackError):
    """Base class for DHCP object errors."""


class DHCPRestartError(DHCPError):
    """Raised when the includes generator on target machine returns non-zero."""


class DHCPConfiguration(ABC):
    """An abstract class which defines the interface for the DHCP configuration generators."""

    def __str__(self) -> str:
        """Return the rendered DHCP configuration snippet."""
        return textwrap.dedent(self._template.format(s=self))

    @property
    @classmethod
    @abstractmethod
    def _template(cls) -> str:
        """Define a string template to be formatted by the instance properties.

        The default implementation of the string representation of the instance will format this template string with
        ``s=self``.

        """

    @property
    @abstractmethod
    def filename(self) -> str:
        """Return a string of the proposed filename for this configuration, from the automation directory."""


@dataclass(frozen=True)
class DHCPConfOpt82(DHCPConfiguration):
    """A configuration generator for host installation DHCP entries via DHCP Option 82.

    Arguments:
        hostname (str): the hostname to generate the DHCP matching block for.
        ipv4 (ipaddress.IPv4Address): the IPv4 to be assigned to the host.
        switch_hostname (str): the hostname of the switch the host is connected to.
        switch_iface (str): the name of the switch interface the host is connected to.
        vlan (str): the name of the VLAN the host is configured for.
        ttys (int): which ttyS to use for this host, accepted values are 0 and 1.
        distro (str): the codename of the Debian distribution to use for the PXE installer.

    """

    hostname: str
    ipv4: IPv4Address
    switch_hostname: str
    switch_iface: str
    vlan: str
    ttys: int
    distro: str

    _template = """
    host {s.hostname} {{
        host-identifier option agent.circuit-id "{s.switch_hostname}:{s.switch_iface}:{s.vlan}";
        fixed-address {s.ipv4};
        option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/{s.distro}-installer/";
    }}
    """

    @property
    def filename(self) -> str:
        """Return the proposed filename based on this configuration."""
        return f"ttyS{self.ttys}-115200/{self.hostname}.conf"


@dataclass(frozen=True)
class DHCPConfMac(DHCPConfiguration):
    """A configuration generator for host installation DHCP entries via MAC address.

    Arguments:
        hostname (str): the hostname to generate the DHCP matching block for.
        ipv4 (ipaddress.IPv4Address): the IPv4 to be assigned to the host.
        mac (str): the MAC address of the host's interface.
        ttys (int): which ttyS to use for this host, accepted values are 0 and 1.
        distro (str): the codename of the Debian distribution to use for the PXE installer.

    """

    hostname: str
    ipv4: IPv4Address
    mac: str
    ttys: int
    distro: str

    _template = """
    host {s.hostname} {{
        hardware ethernet {s.mac};
        fixed-address {s.ipv4};
        option pxelinux.pathprefix "http://apt.wikimedia.org/tftpboot/{s.distro}-installer/";
    }}
    """

    def __post_init__(self) -> None:
        """According to Python's dataclass API to validate the arguments.

        See Also:
            https://docs.python.org/3/library/dataclasses.html#post-init-processing

        """
        mac_pattern = r"[0-9A-Fa-f]{2}(:[0-9A-Fa-f]{2}){5}"
        if re.fullmatch(mac_pattern, self.mac) is None:
            raise DHCPError(f"Got invalid MAC address {self.mac}, not matching pattern {mac_pattern}")

    @property
    def filename(self) -> str:
        """Return the proposed filename based on this configuration."""
        return f"ttyS{self.ttys}-115200/{self.hostname}.conf"


@dataclass(frozen=True)
class DHCPConfMgmt(DHCPConfiguration):
    """A configuration for management network DHCP entries.

    Arguments:
        datacenter (str): the name of the Datacenter the host is.
        serial (str): the vendor serial of the host.
        fqdn (str): the management console FQDN to use for this host.
        ipv4 (ipaddress.IPv4Address): the IP address to give the management interface.

    """

    datacenter: str
    serial: str
    fqdn: str
    ipv4: IPv4Address

    _template = """
    class "{s.fqdn}" {{
        match if (lcase(option host-name) = "idrac-{s.lserial}");
    }}
    pool {{
        allow members of "{s.fqdn}";
        range {s.ipv4} {s.ipv4};
    }}
    """

    def __post_init__(self) -> None:
        """Validate parameters."""
        if self.datacenter not in ALL_DATACENTERS:
            raise DHCPError(f"invalid datacenter {self.datacenter}")
        if not re.search(MGMT_HOSTNAME_RE.format(dc=self.datacenter), self.fqdn):
            raise DHCPError(f"hostname does not look like a valid management hostname: {self.fqdn}")

    @property
    def filename(self) -> str:
        """Return the filename that corresponds to this configuration.

        Returns:
            str: the filename.

        """
        return f"""mgmt-{self.datacenter}/{self.fqdn}.conf"""

    @property
    def lserial(self) -> str:
        """Return the serial as lowercase.

        Returns:
            str: the serial.

        """
        return self.serial.lower()


class DHCP:
    """A class which provides tools for manipulating DHCP configuration snippets by data center."""

    def __init__(self, hosts: RemoteHosts, *, dry_run: bool = True):
        """Create a DHCP instance.

        Arguments:
            hosts (spicerack.remote.RemoteHosts): The target datacenter's install servers.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._dry_run = dry_run
        if len(hosts) < 1:
            raise DHCPError("No target hosts provided.")
        self._hosts = hosts

    def refresh_dhcp(self) -> None:
        """Regenerate includes on target data center and restart DHCP, or raise if failure at any stage."""
        try:
            self._hosts.run_sync("/usr/local/sbin/dhcpincludes -r commit", print_progress_bars=False)
        except RemoteExecutionError as exc:
            raise DHCPRestartError("restarting generating dhcp config or restarting dhcpd failed") from exc

    def push_configuration(self, configuration: DHCPConfiguration) -> None:
        """Push a specified file with specified content to DHCP server and call refresh_dhcp.

        Arguments:
            configuration (spicerack.dhcp.DHCPConfiguration): An instance which provides content and filename for a
                configuration.

        """
        filename = configuration.filename
        try:
            self._hosts.run_sync(
                f"/usr/bin/test '!' '-e'  {DHCP_TARGET_PATH}/{filename}",
                is_safe=True,
                print_output=False,
                print_progress_bars=False,
            )
        except RemoteExecutionError as exc:
            raise DHCPError(f"target file {filename} exists") from exc

        b64encoded = base64.b64encode(str(configuration).encode()).decode()
        try:
            self._hosts.run_sync(
                f"/bin/echo '{b64encoded}' | /usr/bin/base64 -d > {DHCP_TARGET_PATH}/{filename}",
                print_progress_bars=False,
            )
        except RemoteExecutionError as exc:
            raise DHCPError(f"target file {filename} failed to be created.") from exc

        self.refresh_dhcp()

    def remove_configuration(self, configuration: DHCPConfiguration, force: bool = False) -> None:
        """Remove configuration from target DHCP server then call refresh_dhcp.

        This will fail if contents do not match unless force is True.

        Arguments:
            configuration (spicerack.dhcp.DHCPConfiguration): An instance which provides content and filename for a
                                                              configuration.
            force (bool, default False): If set to True, will remove filename regardless.

        """
        if not force:
            confsha256 = sha256(str(configuration).encode()).hexdigest()
            try:
                results = self._hosts.run_sync(
                    f"sha256sum {DHCP_TARGET_PATH}/{configuration.filename}",
                    is_safe=True,
                    print_output=False,
                    print_progress_bars=False,
                )
            except RemoteExecutionError as exc:
                raise DHCPError(f"Can't test {configuration.filename} for removal.") from exc
            seen_match = False
            for _, result in RemoteHosts.results_to_list(results):
                remotesha256 = result.strip().split()[0]
                if remotesha256 != confsha256 and not self._dry_run:
                    raise DHCPError(f"Remote {configuration.filename} has a mismatched SHA256, refusing to remove.")
                seen_match = True
            if not seen_match:
                raise DHCPError("Did not get any result trying to get SHA256, refusing to attempt to remove.")
        try:
            self._hosts.run_sync(
                f"/bin/rm -v {DHCP_TARGET_PATH}/{configuration.filename}", print_output=False, print_progress_bars=False
            )
        except RemoteExecutionError as exc:
            raise DHCPError(f"Can't remove {configuration.filename}.") from exc

        self.refresh_dhcp()

    @contextmanager
    def config(self, dhcp_config: DHCPConfiguration) -> Iterator[None]:
        """A context manager to perform actions while the given DHCP config is valid.

        Arguments:
             dhcp_config (spicerack.dhcp.DHCPConfiguration): The DHCP configuration to use.

        """
        self.push_configuration(dhcp_config)
        try:
            yield
        finally:
            self.remove_configuration(dhcp_config)



"""Ganeti module."""

import logging
from typing import Dict, Optional, Tuple

from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException
from wmflib.requests import http_session

from spicerack.constants import PUPPET_CA_PATH
from spicerack.exceptions import SpicerackError
from spicerack.remote import Remote, RemoteHosts

logger = logging.getLogger(__name__)

RAPI_URL_FORMAT: str = "https://{cluster}:5080"
""":py:class:`str`: the template string to construct the Ganeti RAPI URL."""

CLUSTERS_AND_ROWS: Dict[str, Tuple[str, ...]] = {
    "ganeti01.svc.eqiad.wmnet": ("A", "B", "C", "D"),
    "ganeti01.svc.codfw.wmnet": ("A", "B", "C", "D"),
    "ganeti01.svc.esams.wmnet": ("OE",),
    "ganeti01.svc.ulsfo.wmnet": ("1",),
    "ganeti01.svc.eqsin.wmnet": ("1",),
    "ganeti-test01.svc.codfw.wmnet": ("A",),
    "ganeti01.svc.drmrs.wmnet": ("B12",),
    "ganeti02.svc.drmrs.wmnet": ("B13",),
}
""":py:class:`dict`: the available Ganeti clusters with the set of available rows in each of them."""

INSTANCE_LINKS: Tuple[str, ...] = ("public", "private", "analytics")
""":py:class:`tuple`: the list of possible instance link types."""


class GanetiError(SpicerackError):
    """Raised on errors from Ganeti operations."""


class GanetiRAPI:
    """Class which wraps the read-only Ganeti RAPI."""

    def __init__(self, cluster_url: str, username: str, password: str, timeout: int, ca_path: str):
        """Initialize the instance.

        Arguments:
            cluster_url (str): the URL of the RAPI endpoint.
            username (str): the RAPI user name
            password (str): the RAPI user's password
            timeout (int): the timeout in seconds for each request
            ca_path (str): the path to the signing certificate authority

        """
        self._url = cluster_url
        self._http_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=timeout)
        self._http_session.auth = HTTPBasicAuth(username, password)
        self._http_session.verify = ca_path

    def _api_get_request(self, *targets: str) -> Dict:
        """Perform a RAPI request.

        Arguments:
           *targets: The path components of the request (minus the /2/ part of the path)

        Returns:
           dict: The decoded JSON response

        Raises:
           spicerack.ganeti.GanetiError: on non-200 responses

        """
        full_url = "/".join([self._url, "2"] + list(targets))
        try:
            result = self._http_session.get(full_url)
        except RequestException as ex:
            raise GanetiError("Error while performing request to RAPI") from ex

        if result.status_code != 200:
            raise GanetiError(f"Non-200 from API: {result.status_code}: {result.text}")

        return result.json()

    @property
    def info(self) -> Dict:
        """Return complete cluster information.

        Returns:
            dict: Cluster information dictionary

        Raises:
            spicerack.ganeti.GanetiError: API errors

        """
        return self._api_get_request("info")

    @property
    def master(self) -> Optional[str]:
        """Return the internal name for the current ganeti master node.

        Returns:
            str: The hostname of the master node (or None if the data is missing).

        Raises:
            spicerack.ganeti.GanetiError: API errors

        """
        return self.info.get("master")

    def fetch_instance(self, fqdn: str) -> Dict:
        """Return full information about an instance.

        Arguments:
           fqdn: the FQDN of the instance in question

        Returns:
           dict: host information

        Raises:
           spicerack.ganeti.GanetiError: API errors

        """
        return self._api_get_request("instances", fqdn)

    def fetch_instance_mac(self, fqdn: str) -> str:
        """Convenience method to return the 0th adapter's MAC address for an instance.

        Note that we don't allow the creation of instances with more than one MAC address at this time.

        Arguments:
           fqdn: the FQDN of the instance in question

        Returns:
           str: MAC address

        Raises:
           GanetiError: API errors

        """
        instance_info = self.fetch_instance(fqdn)
        if "nic.macs" not in instance_info or not instance_info["nic.macs"]:
            raise GanetiError("Can't find any MACs for instance")

        return instance_info["nic.macs"][0]


class GntInstance:
    """Class that wraps gnt-instance command execution on a Ganeti cluster master host."""

    def __init__(self, master: RemoteHosts, cluster: str, instance: str):
        """Initialize the instance.

        Arguments:
            master (spicerack.remote.RemoteHosts): the Ganeti cluster master remote instance.
            cluster (str): the Ganeti cluster name.
            instance (str): the FQDN of the Ganeti VM instance to act upon.

        """
        self._master = master
        self._cluster = cluster
        self._instance = instance

    @property
    def cluster(self) -> str:
        """Getter for the Ganeti cluster property.

        Returns:
            str: the Ganeti cluster name the instance belongs to.

        """
        return self._cluster

    def shutdown(self, *, timeout: int = 2) -> None:
        """Shutdown the Ganeti VM instance.

        Arguments:
            timeout (int): time in minutes to wait for a clean shutdown before pulling the plug.

        """
        logger.info("Shutting down VM %s in cluster %s", self._instance, self._cluster)
        self._master.run_sync(f"gnt-instance shutdown --force --timeout={timeout} {self._instance}")

    def remove(self, *, shutdown_timeout: int = 2) -> None:
        """Shutdown and remove the VM instance from the Ganeti cluster, including its disks.

        Arguments:
            shutdown_timeout (int): time in minutes to wait for a clean shutdown before pulling the plug.

        Note:
            This action requires few minutes, inform the user about the waiting time when using this method.

        """
        logger.info(
            "Removing VM %s in cluster %s. This may take a few minutes.",
            self._instance,
            self._cluster,
        )
        self._master.run_sync(f"gnt-instance remove --shutdown-timeout={shutdown_timeout} --force {self._instance}")

    def add(self, *, row: str, vcpus: int, memory: int, disk: int, link: str) -> None:
        """Create the VM for the instance in the Ganeti cluster with the specified characteristic.

        Arguments:
            row (str): the Datacenter physical row where to allocate the instance, one of
                :py:const:`spicerack.ganeti.CLUSTERS_AND_ROWS` based on the current cluster.
            vcpus (int): the number of virtual CPUs to assign to the instance.
            memory (int): the amount of RAM to assign to the instance in gigabytes.
            disk (int): the amount of disk to assign to the instance in gigabytes.
            link (str): the type of network link to use, one of :py:const:`spicerack.ganeti.INSTANCE_LINKS`.

        Raises:
            spicerack.ganeti.GanetiError: on parameter validation error.

        Note:
            This action requires few minutes, inform the user about the waiting time when using this method.

        """
        if link not in INSTANCE_LINKS:
            raise GanetiError(f"Invalid link '{link}', expected one of: {INSTANCE_LINKS}")

        if row not in CLUSTERS_AND_ROWS[self._cluster]:
            raise GanetiError(
                f"Invalid row '{row}' for cluster {self._cluster}, expected one of: {CLUSTERS_AND_ROWS[self._cluster]}"
            )

        local_vars = locals()
        for var_label in ("vcpus", "memory", "disk"):
            if local_vars[var_label] <= 0:
                raise GanetiError(
                    f"Invalid value '{local_vars[var_label]}' for {var_label}, expected positive integer."
                )

        command = (
            "gnt-instance add"
            " -t drbd"
            " -I hail"
            f" --net 0:link={link}"
            " --hypervisor-parameters=kvm:boot_order=network"
            " -o debootstrap+default"
            " --no-install"
            f" -g row_{row}"
            f" -B vcpus={vcpus},memory={memory}g"
            f" --disk 0:size={disk}g"
            f" {self._instance}"
        )

        logger.info(
            (
                "Creating VM %s in cluster %s with row=%s vcpus=%d memory=%dGB disk=%dGB link=%s. "
                "This may take a few minutes."
            ),
            self._instance,
            self._cluster,
            row,
            vcpus,
            memory,
            disk,
            link,
        )

        results = self._master.run_sync(command)
        for _, output in results:
            logger.info(output.message().decode())


class Ganeti:
    """Class which wraps all Ganeti clusters."""

    def __init__(self, username: str, password: str, timeout: int, remote: Remote):
        """Initialize the instance.

        Arguments:
            username (str): The RAPI username to use.
            password (str): The RAPI password to use.
            timeout (int): The timeout in seconds for each request to the API.
            remote (spicerack.remote.Remote): the remote instance to connect to Ganeti hosts.

        """
        self._username = username
        self._password = password
        self._timeout = timeout
        self._remote = remote

    def rapi(self, cluster: str) -> GanetiRAPI:
        """Return a RAPI object for a particular cluster.

        Arguments:
            cluster (str): the name of the Ganeti cluster to get a RAPI for.

        Returns:
            spicerack.ganeti.GanetiRAPI: the RAPI interface object

        Raises:
            spicerack.ganeti.GanetiError: on an invalid cluster name

        """
        if cluster not in CLUSTERS_AND_ROWS:
            keys = tuple(CLUSTERS_AND_ROWS.keys())
            raise GanetiError(f"Cannot find cluster {cluster} (expected {keys}).")

        cluster_url = RAPI_URL_FORMAT.format(cluster=cluster)

        return GanetiRAPI(cluster_url, self._username, self._password, self._timeout, PUPPET_CA_PATH)

    def fetch_cluster_for_instance(self, fqdn: str) -> str:
        """Return the cluster name for a given FQDN if possible.

        Arguments:
            fqdn (str): The FQDN for the host to locate.

        Returns:
            str: The cluster name if found.

        Raises:
           spicerack.ganeti.GanetiError: if the host was not found in any configured cluster.

        """
        for cluster in CLUSTERS_AND_ROWS:
            cluster_rapi = self.rapi(cluster)
            try:
                cluster_rapi.fetch_instance(fqdn)
                return cluster
            except GanetiError:
                continue

        raise GanetiError(f"Cannot find {fqdn} in any configured cluster.")

    def instance(self, instance: str, *, cluster: str = "") -> GntInstance:
        """Return an instance of GntInstance to perform RW operation on the given Ganeti VM instance.

        Arguments:
            instance (str): the FQDN of the Ganeti VM instance to act upon.
            cluster (str, optional): the name of the Ganeti cluster where to look for the instance.

        Returns:
            spicerack.ganeti.GntInstance: ready to perform RW actions.

        """
        if not cluster:
            cluster = self.fetch_cluster_for_instance(instance)
        master = self.rapi(cluster).master
        if master is None:
            raise GanetiError(f"Master for cluster {cluster} is None")

        return GntInstance(self._remote.query(master), cluster, instance)



"""Debmonitor module tests."""
import pytest
import requests

from spicerack import debmonitor

DEBMONITOR_HOST = "debmonitor.example.com"
HOST1_URL = f"https://{DEBMONITOR_HOST}/hosts/host1.example.com"


class TestDebmonitor:
    """Debmonitor class tests."""

    def setup_method(self):
        """Initialize the test environment for Debmonitor."""
        # pylint: disable=attribute-defined-outside-init
        self.debmonitor = debmonitor.Debmonitor(DEBMONITOR_HOST, "cert.pem", "key.pem", dry_run=False)
        self.debmonitor_dry_run = debmonitor.Debmonitor(DEBMONITOR_HOST, "cert.pem", "key.pem")

    def test_host_delete_ok(self, requests_mock):
        """It should delete the host from Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes["no_content"])
        self.debmonitor.host_delete("host1.example.com")
        assert requests_mock.call_count == 1
        assert requests_mock.request_history[0].method == "DELETE"
        assert requests_mock.request_history[0].url == HOST1_URL

    def test_host_delete_dry_run(self, requests_mock):
        """It should not delete the host from Debmonitor in DRY-RUN."""
        self.debmonitor_dry_run.host_delete("host1.example.com")
        assert not requests_mock.called

    def test_host_delete_not_found(self, requests_mock):
        """It should not raise if the host is already absent in Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes["not_found"])
        self.debmonitor.host_delete("host1.example.com")
        assert requests_mock.call_count == 1

    def test_host_delete_fail(self, requests_mock):
        """It should raise DebmonitorError if failing to delete the host from Debmonitor."""
        requests_mock.delete(HOST1_URL, status_code=requests.codes["bad_request"])
        with pytest.raises(
            debmonitor.DebmonitorError,
            match="Unable to remove host host1.example.com from Debmonitor",
        ):
            self.debmonitor.host_delete("host1.example.com")

        assert requests_mock.call_count == 1



"""Administrative module tests."""
import pytest

from spicerack import administrative


def test_reason_init():
    """It should initalize a Reason instance with the mandatory parameters."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert str(reason) == "Reason message - user1@host1"


def test_reason_init_with_task():
    """It should initalize a Reason instance with the optional task ID."""
    reason = administrative.Reason("Reason message", "user1", "host1", task_id="T12345")
    assert str(reason) == "Reason message - user1@host1 - T12345"


@pytest.mark.parametrize(
    "failing_param_name, args, kwargs",
    (
        ("_reason", ['Reason with "double quotes"', "user1", "host1"], {"task_id": ""}),
        ("_username", ["Reason message", 'user"', "host1"], {"task_id": ""}),
        ("_hostname", ["Reason message", "user1", 'host1"'], {"task_id": ""}),
        ("_task_id", ["Reason message", "user1", "host1"], {"task_id": 'T"'}),
    ),
)
def test_reason_init_double_quotes_in_params(failing_param_name, args, kwargs):
    """It should raise ReasonError if any of the parameters contains double quotes."""
    with pytest.raises(
        administrative.ReasonError,
        match=f"Property {failing_param_name} cannot contain double quotes",
    ):
        administrative.Reason(*args, **kwargs)


def test_reason_reason():
    """It should return the reason."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert reason.reason == "Reason message"


def test_reason_owner():
    """It should return the owner part."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert reason.owner == "user1@host1"


def test_reason_hostname():
    """It should return the hostname."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert reason.hostname == "host1"


def test_reason_quoted():
    """It should return the double quoted string representation of the instance."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert reason.quoted() == '"Reason message - user1@host1"'


def test_reason_task_id_without_task():
    """It should return None for the missing task ID."""
    reason = administrative.Reason("Reason message", "user1", "host1")
    assert reason.task_id is None


def test_reason_task_id_with_task():
    """It should return the task ID."""
    reason = administrative.Reason("Reason message", "user1", "host1", task_id="T12345")
    assert reason.task_id == "T12345"
